@@ -60,6 +60,16 @@ void Session_sysvars_tracker::vars_list::copy(vars_list* from, THD *thd)
   from->init();
 }
 
+Session_sysvars_tracker::
+sysvar_node_st *Session_sysvars_tracker::vars_list::search(const sys_var *svar)
+{
+  return reinterpret_cast<sysvar_node_st*>(
+           my_hash_search(&m_registered_sysvars,
+                         reinterpret_cast<const uchar*>(&svar->offset),
+                         sizeof(svar->offset)));
+}
+
+
 /**
   Inserts the variable to be tracked into m_registered_sysvars hash.
 
@@ -167,7 +177,7 @@ bool Session_sysvars_tracker::vars_list::parse_var_list(THD *thd,
     {
       push_warning_printf(thd, Sql_condition::WARN_LEVEL_WARN,
                           ER_WRONG_VALUE_FOR_VAR,
-                          "%.*s is not a valid system variable and will"
+                          "%.*s is not a valid system variable and will "
                           "be ignored.", (int)var.length, token);
     }
     else
@@ -221,7 +231,7 @@ bool sysvartrack_validate_value(THD *thd, const char *str, size_t len)
     /* Remove leading/trailing whitespace. */
     trim_whitespace(system_charset_info, &var);
 
-    if (!strcmp(var.str, "*") && !find_sys_var(thd, var.str, var.length))
+    if (strcmp(var.str, "*") && !find_sys_var(thd, var.str, var.length))
       return true;
 
     if (lasts)
@@ -331,10 +341,9 @@ void Session_sysvars_tracker::init(THD *thd)
   mysql_mutex_assert_owner(&LOCK_global_system_variables);
   DBUG_ASSERT(thd->variables.session_track_system_variables ==
               global_system_variables.session_track_system_variables);
-  DBUG_ASSERT(global_system_variables.session_track_system_variables);
   thd->variables.session_track_system_variables=
     my_strdup(PSI_INSTRUMENT_ME,
-              global_system_variables.session_track_system_variables,
+              safe_str(global_system_variables.session_track_system_variables),
               MYF(MY_WME | MY_THREAD_SPECIFIC));
 }
 
@@ -429,6 +438,7 @@ bool Session_sysvars_tracker::vars_list::store(THD *thd, String *buf)
     SHOW_VAR show;
     CHARSET_INFO *charset;
     size_t val_length, length;
+    mysql_mutex_lock(&LOCK_global_system_variables);
     mysql_mutex_lock(&LOCK_plugin);
     if (!*node->test_load)
     {
@@ -445,13 +455,12 @@ bool Session_sysvars_tracker::vars_list::store(THD *thd, String *buf)
     show.name= svar->name.str;
     show.value= (char *) svar;
 
-    mysql_mutex_lock(&LOCK_global_system_variables);
     const char *value= get_one_variable(thd, &show, OPT_SESSION, SHOW_SYS, NULL,
                                         &charset, val_buf, &val_length);
-    mysql_mutex_unlock(&LOCK_global_system_variables);
 
     if (is_plugin)
       mysql_mutex_unlock(&LOCK_plugin);
+    mysql_mutex_unlock(&LOCK_global_system_variables);
 
     length= net_length_size(svar->name.length) +
       svar->name.length +
@@ -560,8 +569,9 @@ uchar *Session_sysvars_tracker::sysvars_get_key(const char *entry,
                                                 size_t *length,
                                                 my_bool not_used __attribute__((unused)))
 {
-  *length= sizeof(sys_var *);
-  return (uchar *) &(((sysvar_node_st *) entry)->m_svar);
+  auto key=&(((sysvar_node_st *) entry)->m_svar->offset);
+  *length= sizeof(*key);
+  return (uchar *) key;
 }
 
 
@@ -576,6 +586,12 @@ bool sysvartrack_global_update(THD *thd, char *str, size_t len)
 {
   LEX_STRING tmp= { str, len };
   Session_sysvars_tracker::vars_list dummy;
+  DBUG_EXECUTE_IF("dbug_session_tracker_parse_error",
+                  {
+                    my_error(ER_OUTOFMEMORY, MYF(0), 1);
+                    return true;
+                  });
+
   if (!dummy.parse_var_list(thd, tmp, false, system_charset_info))
   {
     dummy.construct_var_list(str, len + 1);
@@ -996,7 +1012,7 @@ enum_tx_state Transaction_state_tracker::calc_trx_state(THD *thd,
                                                         bool has_trx)
 {
   enum_tx_state      s;
-  bool               read= (l <= TL_READ_NO_INSERT);
+  bool               read= (l < TL_FIRST_WRITE);
 
   if (read)
     s= has_trx ? TX_READ_TRX  : TX_READ_UNSAFE;
