@@ -64,7 +64,6 @@ Created 2/16/1996 Heikki Tuuri
 #include "rem0rec.h"
 #include "srv0start.h"
 #include "srv0srv.h"
-#include "btr0defragment.h"
 #include "mysql/service_wsrep.h" /* wsrep_recovery */
 #include "trx0rseg.h"
 #include "buf0flu.h"
@@ -175,7 +174,7 @@ static dberr_t create_log_file(bool create_new_db, lsn_t lsn)
 
 	/* We will retain ib_logfile0 until we have written a new logically
 	empty log as ib_logfile101 and atomically renamed it to
-	ib_logfile0 in log_t::rename_resized(). */
+	ib_logfile0 in log_t::resize_rename(). */
 	delete_log_files();
 
 	ut_ad(!os_aio_pending_reads());
@@ -222,7 +221,12 @@ close_and_exit:
 	if (!log_sys.attach(file, srv_log_file_size)) {
 		goto close_and_exit;
 	}
-	if (!fil_system.sys_space->open(create_new_db)) {
+
+	mysql_mutex_lock(&recv_sys.mutex);
+	const bool all_opened = fil_system.sys_space->open(create_new_db);
+	mysql_mutex_unlock(&recv_sys.mutex);
+
+	if (!all_opened) {
 		goto err_exit;
 	}
 
@@ -334,6 +338,15 @@ inline dberr_t trx_sys_t::reset_page(mtr_t *mtr)
 
   if (!sys_header) return err;
 
+  if (mach_read_from_4(sys_header->page.frame + TRX_SYS +
+                       TRX_SYS_FSEG_HEADER) != TRX_SYS_SPACE)
+    return DB_CORRUPTION;
+
+  /* Store the TRX_SYS_FSEG_HEADER page, offset */
+  char fseg_addr[6];
+  memcpy(fseg_addr,
+         sys_header->page.frame + TRX_SYS + TRX_SYS_FSEG_HEADER + 4, 6);
+
   const bool dblwr_enabled=
     mach_read_from_4(TRX_SYS_DOUBLEWRITE_MAGIC + TRX_SYS_DOUBLEWRITE +
                      sys_header->page.frame)
@@ -348,6 +361,17 @@ inline dberr_t trx_sys_t::reset_page(mtr_t *mtr)
   mtr->write<2>(*sys_header, FIL_PAGE_TYPE + sys_header->page.frame,
                 FIL_PAGE_TYPE_TRX_SYS);
 
+  DBUG_EXECUTE_IF("sys_fseg_header_fail",
+                  {
+                    fseg_addr[4]= 0;
+                    fseg_addr[5]= 0;
+                  });
+
+  /** Write the TRX_SYS_FSEG_HEADER only if it is not zero-filled */
+  if (!memcmp(fseg_addr + 4, "00", 2))
+    mtr->memcpy(*sys_header,
+                sys_header->page.frame + TRX_SYS + TRX_SYS_FSEG_HEADER + 4,
+                fseg_addr, 6);
   mtr->write<4>(*sys_header,
                 TRX_SYS + TRX_SYS_RSEGS + TRX_SYS_RSEG_PAGE_NO +
                 sys_header->page.frame, FSP_FIRST_RSEG_PAGE_NO);
@@ -370,6 +394,11 @@ inline dberr_t trx_sys_t::reset_page(mtr_t *mtr)
       sys_header->page.frame + TRX_SYS_DOUBLEWRITE
       + FSEG_HEADER_SIZE + TRX_SYS_DOUBLEWRITE_REPEAT,
       sys_header->page.frame + TRX_SYS_DOUBLEWRITE + FSEG_HEADER_SIZE, 12);
+    mtr->write<4>(
+      *sys_header,
+      TRX_SYS_DOUBLEWRITE + TRX_SYS_DOUBLEWRITE_SPACE_ID_STORED +
+      sys_header->page.frame,
+      TRX_SYS_DOUBLEWRITE_SPACE_ID_STORED_N);
   }
 
   return DB_SUCCESS;
@@ -637,28 +666,41 @@ static uint32_t srv_undo_tablespace_open(bool create, const char* name,
   if (!success)
     return 0;
 
+  ulint n_retries = 5;
   os_offset_t size= os_file_get_size(fh);
   ut_a(size != os_offset_t(-1));
+  page_t *apage= nullptr;
+  const page_t *page= nullptr;
 
   if (!create)
   {
-    page_t *page= static_cast<byte*>(aligned_malloc(srv_page_size,
-                                                    srv_page_size));
-    if (os_file_read(IORequestRead, fh, page, 0, srv_page_size, nullptr) !=
+    apage= static_cast<byte*>(aligned_malloc(srv_page_size, srv_page_size));
+undo_retry:
+    if (os_file_read(IORequestRead, fh, apage, 0, srv_page_size, nullptr) !=
         DB_SUCCESS)
     {
 err_exit:
-      ib::error() << "Unable to read first page of file " << name;
-      aligned_free(page);
+      if (n_retries && srv_operation == SRV_OPERATION_BACKUP)
+      {
+        sql_print_information("InnoDB: Retrying to read undo "
+                              "tablespace %s", name);
+        n_retries--;
+        goto undo_retry;
+      }
+      sql_print_error("InnoDB: Unable to read first page of file %s", name);
+      aligned_free(apage);
       return ~0U;
     }
 
+    page= apage;
+    DBUG_EXECUTE_IF("undo_space_read_fail", goto err_exit;);
+
     uint32_t id= mach_read_from_4(FIL_PAGE_SPACE_ID + page);
     if (id == 0 || id >= SRV_SPACE_ID_UPPER_BOUND ||
         memcmp_aligned<2>(FIL_PAGE_SPACE_ID + page,
                           FSP_HEADER_OFFSET + FSP_SPACE_ID + page, 4))
     {
-      ib::error() << "Inconsistent tablespace ID in file " << name;
+      sql_print_error("InnoDB: Inconsistent tablespace ID in file %s", name);
       goto err_exit;
     }
 
@@ -667,13 +709,21 @@ err_exit:
 
     if (buf_page_is_corrupted(false, page, fsp_flags))
     {
-      sql_print_error("InnoDB: Checksum mismatch in the first page of file %s",
-                      name);
-      if (recv_sys.dblwr.restore_first_page(space_id, name, fh))
+      page= recv_sys.dblwr.find_page(page_id_t{space_id, 0}, LSN_MAX);
+      if (!page)
+      {
+        /* If the first page of the given user tablespace is not there
+        in the doublewrite buffer, then the recovery is going to fail
+        now. Report error only when doublewrite buffer is not empty */
+        sql_print_error("InnoDB: Corrupted page "
+                        "[page id: space=" UINT32PF ", page number=0]"
+                        " of datafile '%s' could not be found"
+                        " in the doublewrite buffer", space_id, name);
         goto err_exit;
-    }
+      }
 
-    aligned_free(page);
+      fsp_flags= mach_read_from_4(FSP_HEADER_OFFSET + FSP_SPACE_FLAGS + page);
+    }
   }
 
   /* Load the tablespace into InnoDB's internal data structures. */
@@ -696,7 +746,7 @@ err_exit:
     space->set_sizes(SRV_UNDO_TABLESPACE_SIZE_IN_PAGES);
     space->size= file->size= uint32_t(size >> srv_page_size_shift);
   }
-  else if (!file->read_page0())
+  else if (!file->read_page0(page, true))
   {
     os_file_close(file->handle);
     file->handle= OS_FILE_CLOSED;
@@ -705,6 +755,7 @@ err_exit:
   }
 
   mysql_mutex_unlock(&fil_system.mutex);
+  aligned_free(apage);
   return space_id;
 }
 
@@ -812,7 +863,7 @@ unused_undo:
   {
      char name[OS_FILE_MAX_PATH];
      snprintf(name, sizeof name, "%s/undo%03u", srv_undo_dir, i);
-     uint32_t space_id= srv_undo_tablespace_open(create_new_undo, name, i);
+     uint32_t space_id= srv_undo_tablespace_open(false, name, i);
      if (!space_id || space_id == ~0U)
        break;
      if (0 == srv_undo_tablespaces_open++)
@@ -863,11 +914,12 @@ dberr_t srv_undo_tablespaces_init(bool create_new_undo, mtr_t *mtr)
   srv_undo_tablespaces_active= srv_undo_tablespaces;
 
   uint32_t n_undo= (create_new_undo || srv_operation == SRV_OPERATION_BACKUP ||
-                 srv_operation == SRV_OPERATION_RESTORE_DELTA)
+                    srv_operation == SRV_OPERATION_RESTORE_DELTA)
     ? srv_undo_tablespaces : TRX_SYS_N_RSEGS;
 
-  if (dberr_t err= srv_all_undo_tablespaces_open(create_new_undo, n_undo))
-    return err;
+  mysql_mutex_lock(&recv_sys.mutex);
+  dberr_t err= srv_all_undo_tablespaces_open(create_new_undo, n_undo);
+  mysql_mutex_unlock(&recv_sys.mutex);
 
   /* Initialize srv_undo_space_id_start=0 when there are no
   dedicated undo tablespaces. */
@@ -875,16 +927,11 @@ dberr_t srv_undo_tablespaces_init(bool create_new_undo, mtr_t *mtr)
     srv_undo_space_id_start= 0;
 
   if (create_new_undo)
-  {
-    for (uint32_t i= 0; i < srv_undo_tablespaces; ++i)
-    {
-      dberr_t err= fsp_header_init(fil_space_get(srv_undo_space_id_start + i),
-                                   SRV_UNDO_TABLESPACE_SIZE_IN_PAGES, mtr);
-      if (err) return err;
-    }
-  }
+    for (uint32_t i= 0; err == DB_SUCCESS && i < srv_undo_tablespaces; ++i)
+      err= fsp_header_init(fil_space_get(srv_undo_space_id_start + i),
+                           SRV_UNDO_TABLESPACE_SIZE_IN_PAGES, mtr);
 
-  return DB_SUCCESS;
+  return err;
 }
 
 /** Create the temporary file tablespace.
@@ -1083,7 +1130,8 @@ same_size:
 
   log_sys.latch.wr_unlock();
 
-  log_write_up_to(flushed_lsn, false);
+  if (latest_format)
+    log_write_up_to(flushed_lsn, false);
 
   ut_ad(flushed_lsn == log_sys.get_lsn());
   ut_ad(!os_aio_pending_reads());
@@ -1152,8 +1200,11 @@ static dberr_t srv_log_rebuild_if_needed()
   return srv_log_rebuild();
 }
 
+ut_d(bool ibuf_upgrade_was_needed;)
+
 ATTRIBUTE_COLD static dberr_t ibuf_log_rebuild_if_needed()
 {
+  ut_d(ibuf_upgrade_was_needed= true;)
   mysql_mutex_lock(&recv_sys.mutex);
   recv_sys.apply(true);
   mysql_mutex_unlock(&recv_sys.mutex);
@@ -1341,10 +1392,7 @@ dberr_t srv_start(bool create_new_db)
 	}
 #endif /* UNIV_DEBUG */
 
-	if (!log_sys.create()) {
-		return srv_init_abort(DB_ERROR);
-	}
-
+	log_sys.create();
 	recv_sys.create();
 	lock_sys.create(srv_lock_table_size);
 
@@ -1419,9 +1467,14 @@ dberr_t srv_start(bool create_new_db)
 
 	/* Open data files in the system tablespace: we keep
 	them open until database shutdown */
+	mysql_mutex_lock(&recv_sys.mutex);
 	ut_d(fil_system.sys_space->recv_size = srv_sys_space_size_debug);
+	err = fil_system.sys_space->open(create_new_db)
+		? DB_SUCCESS : DB_ERROR;
+	mysql_mutex_unlock(&recv_sys.mutex);
 
-	if (fil_system.sys_space->open(create_new_db)) {
+
+	if (err == DB_SUCCESS) {
 		mtr_t mtr;
 		mtr.start();
 		err= srv_undo_tablespaces_init(create_new_db, &mtr);
@@ -1767,6 +1820,13 @@ dberr_t srv_start(bool create_new_db)
 		ut_ad(high_level_read_only
 		      || srv_force_recovery < SRV_FORCE_NO_UNDO_LOG_SCAN);
 
+		if (!high_level_read_only
+		    && srv_sys_space.can_auto_shrink()) {
+			fsp_system_tablespace_truncate(false);
+			DBUG_EXECUTE_IF("crash_after_sys_truncate",
+					return srv_init_abort(DB_ERROR););
+		}
+
 		/* Validate a few system page types that were left
 		uninitialized before MySQL or MariaDB 5.5. */
 		if (!high_level_read_only
@@ -1905,13 +1965,13 @@ skip_monitors:
 	if (srv_print_verbose_log) {
 		sql_print_information("InnoDB: "
 				      "log sequence number " LSN_PF
-#ifdef HAVE_PMEM
+#ifdef HAVE_INNODB_MMAP
 				      "%s"
 #endif
 				      "; transaction id " TRX_ID_FMT,
 				      recv_sys.lsn,
-#ifdef HAVE_PMEM
-				      log_sys.is_pmem()
+#ifdef HAVE_INNODB_MMAP
+				      log_sys.is_mmap()
 				      ? " (memory-mapped)" : "",
 #endif
 				      trx_sys.get_max_trx_id());
@@ -1947,9 +2007,6 @@ skip_monitors:
 		fil_crypt_threads_cond. */
 		fil_crypt_threads_init();
 
-		/* Initialize online defragmentation. */
-		btr_defragment_init();
-
 		srv_started_redo = true;
 	}
 
@@ -2044,7 +2101,6 @@ void innodb_shutdown()
 		fts_optimize_shutdown(); dict_stats_shutdown(); */
 
 		fil_crypt_threads_cleanup();
-		btr_defragment_shutdown();
 	}
 
 	/* This must be disabled before closing the buffer pool
