@@ -152,9 +152,7 @@ trx_undo_log_v_idx(
 	ulint n_idx = 0;
 	for (const auto& v_index : vcol->v_indexes) {
 		n_idx++;
-		if (uint32_t hi= uint32_t(v_index.index->id >> 32)) {
-			size += 1 + mach_get_compressed_size(hi);
-		}
+		/* FIXME: index->id is 64 bits! */
 		size += mach_get_compressed_size(uint32_t(v_index.index->id));
 		size += mach_get_compressed_size(v_index.nth_field);
 	}
@@ -181,14 +179,10 @@ trx_undo_log_v_idx(
 	ptr += mach_write_compressed(ptr, n_idx);
 
 	for (const auto& v_index : vcol->v_indexes) {
-		/* This is compatible with
-		ptr += mach_u64_write_much_compressed(ptr, v_index.index-id)
-		(the added "if" statement is fixing an old regression). */
-		if (uint32_t hi= uint32_t(v_index.index->id >> 32)) {
-			*ptr++ = 0xff;
-			ptr += mach_write_compressed(ptr, hi);
-		}
-		ptr += mach_write_compressed(ptr, uint32_t(v_index.index->id));
+		ptr += mach_write_compressed(
+			/* FIXME: index->id is 64 bits! */
+			ptr, uint32_t(v_index.index->id));
+
 		ptr += mach_write_compressed(ptr, v_index.nth_field);
 	}
 
@@ -227,15 +221,7 @@ trx_undo_read_v_idx_low(
 	dict_index_t*	clust_index = dict_table_get_first_index(table);
 
 	for (ulint i = 0; i < num_idx; i++) {
-		index_id_t	id = 0;
-		/* This is like mach_u64_read_much_compressed(),
-		but advancing ptr to the next field. */
-		if (*ptr == 0xff) {
-			ptr++;
-			id = mach_read_next_compressed(&ptr);
-			id <<= 32;
-		}
-		id |= mach_read_next_compressed(&ptr);
+		index_id_t	id = mach_read_next_compressed(&ptr);
 		ulint		pos = mach_read_next_compressed(&ptr);
 		dict_index_t*	index = dict_table_get_next_index(clust_index);
 
@@ -1773,8 +1759,7 @@ TRANSACTIONAL_TARGET ATTRIBUTE_NOINLINE
 /** @return whether the transaction holds an exclusive lock on a table */
 static bool trx_has_lock_x(const trx_t &trx, dict_table_t& table)
 {
-  if (table.is_temporary())
-    return true;
+  ut_ad(!table.is_temporary());
 
   uint32_t n;
 
@@ -1855,7 +1840,7 @@ trx_undo_report_row_operation(
 	auto m = trx->mod_tables.emplace(index->table, trx->undo_no);
 	ut_ad(m.first->second.valid(trx->undo_no));
 
-	if (m.second && index->table->is_active_ddl()) {
+	if (m.second && index->table->is_native_online_ddl()) {
 		trx->apply_online_log= true;
 	}
 
@@ -1875,9 +1860,17 @@ trx_undo_report_row_operation(
 		ut_ad(que_node_get_type(thr->run_node) == QUE_NODE_INSERT);
 		ut_ad(trx->bulk_insert);
 		return DB_SUCCESS;
-	} else if (m.second && trx->bulk_insert
-		   && trx_has_lock_x(*trx, *index->table)) {
-		m.first->second.start_bulk_insert();
+	} else if (!m.second || !trx->bulk_insert) {
+		bulk = false;
+	} else if (index->table->is_temporary()) {
+	} else if (trx_has_lock_x(*trx, *index->table)
+		   && index->table->bulk_trx_id == trx->id) {
+		m.first->second.start_bulk_insert(index->table);
+
+		if (dberr_t err = m.first->second.bulk_insert_buffered(
+			    *clust_entry, *index, trx)) {
+			return err;
+		}
 	} else {
 		bulk = false;
 	}
@@ -2054,10 +2047,8 @@ err_exit:
 
 static dberr_t trx_undo_prev_version(const rec_t *rec, dict_index_t *index,
                                      rec_offs *offsets, mem_heap_t *heap,
-                                     rec_t **old_vers,
-                                     const purge_sys_t::view_guard &check,
-                                     ulint v_status,
-                                     mem_heap_t *v_heap, dtuple_t **vrow,
+                                     rec_t **old_vers, mem_heap_t *v_heap,
+                                     dtuple_t **vrow, ulint v_status,
                                      const trx_undo_rec_t *undo_rec);
 
 inline const buf_block_t *
@@ -2158,8 +2149,7 @@ dberr_t trx_undo_prev_version_build(const rec_t *rec, dict_index_t *index,
       if (UNIV_UNLIKELY(end > offset &&
                         end < srv_page_size - FIL_PAGE_DATA_END))
         err= trx_undo_prev_version(rec, index, offsets, heap,
-                                   old_vers, check, v_status, v_heap, vrow,
-                                   undo_rec);
+                                   old_vers, v_heap, vrow, v_status, undo_rec);
     }
   }
 
@@ -2169,10 +2159,8 @@ dberr_t trx_undo_prev_version_build(const rec_t *rec, dict_index_t *index,
 
 static dberr_t trx_undo_prev_version(const rec_t *rec, dict_index_t *index,
                                      rec_offs *offsets, mem_heap_t *heap,
-                                     rec_t **old_vers,
-                                     const purge_sys_t::view_guard &check,
-                                     ulint v_status,
-                                     mem_heap_t *v_heap, dtuple_t **vrow,
+                                     rec_t **old_vers, mem_heap_t *v_heap,
+                                     dtuple_t **vrow, ulint v_status,
                                      const trx_undo_rec_t *undo_rec)
 {
 	byte type, cmpl_info;
@@ -2242,7 +2230,7 @@ static dberr_t trx_undo_prev_version(const rec_t *rec, dict_index_t *index,
 		the BLOB. */
 
 		if (update->info_bits & REC_INFO_DELETED_FLAG
-		    && check.view().changes_visible(trx_id)) {
+		    && purge_sys.is_purgeable(trx_id)) {
 			return DB_SUCCESS;
 		}
 
